#include <stdio.h>
#include <string.h>

#include <glib.h>

#include "gmime-part.h"
#include "gmime-content.h"
#include "gmime-utils.h"

/**
***  Shared
**/

static void
g_mime_content_append_encoded_impl (GMimeContent* content,
                                    const guchar * encoded,
                                    guint encoded_len,
                                    GMimePartEncodingType encoding,
                                    guchar ** setme_decoded,
                                    guint * setme_decoded_len)
{
	gint len = 0;
	gchar * buf = g_new0 (gchar, encoded_len);

	g_return_if_fail (content!=NULL);
	g_return_if_fail (encoded!=NULL);
	g_return_if_fail (encoded_len>0);
	g_return_if_fail (setme_decoded!=NULL);
	g_return_if_fail (setme_decoded_len!=NULL);

        /* decode & write */
	switch (encoding)
	{
		case GMIME_PART_ENCODING_BASE64:
			len = g_mime_utils_base64_decode_step (encoded, encoded_len,
			                                       buf,
			                                       &content->append_state, &content->append_save);
			break;

		case GMIME_PART_ENCODING_QUOTEDPRINTABLE:
			len = g_mime_utils_quoted_decode_step (encoded, encoded_len,
			                                       buf,
			                                       &content->append_state, &content->append_save);
			break;

		default:
			len = encoded_len;
			memcpy (buf, encoded, encoded_len);

			/* do some smart 8bit detection */
			if (encoding == GMIME_PART_ENCODING_DEFAULT) {
				/* if we were at default before and this is 8bit text, set encoding to 8bit */
				if (content->encoding==GMIME_PART_ENCODING_DEFAULT && g_mime_utils_text_is_8bit (buf, len))
					encoding = GMIME_PART_ENCODING_8BIT;

				/* if we guessed 8bit before but this isn't an 8bit piece, set encoding back to default */
				else if (content->encoding == GMIME_PART_ENCODING_8BIT)
					encoding = g_mime_utils_text_is_8bit (buf, len)
						? GMIME_PART_ENCODING_8BIT : GMIME_PART_ENCODING_DEFAULT;
			}
	}

	/*FIXME*/
	content->encoding = encoding;
	*setme_decoded = buf;
	*setme_decoded_len = len;
}

/**
***  File
**/

typedef struct
{
	gchar * filename;
	FILE * fp;
}
GMimeContentFileData;

static void
g_mime_content_file_data_destroy (gpointer data)
{
	GMimeContentFileData * d = (GMimeContentFileData*) data;
	if (d->fp != NULL)
		fclose (d->fp);
	if (d->filename != NULL) {
		remove (d->filename);
		g_free (d->filename);
	}

	g_free (d);
}

static gint tmp_counter = 0;

static void
g_mime_content_file_init_filename (GMimeContentFileData * data, const gchar * open_args)
{
	/* close old */
	if (data->filename != NULL) {
		if (data->fp != NULL) {
			fclose (data->fp);
			data->fp = NULL;
		}
		remove (data->filename);
		g_free (data->filename);
	}

	/* FIXME */
	data->filename = g_strdup_printf ("/tmp/gmime_test_%d", ++tmp_counter);

	if (open_args != NULL) {
		data->fp = fopen (data->filename, open_args);
		g_return_if_fail (data->fp!=NULL);
	}
}

static void
g_mime_content_file_set_begin (GMimeContent* content)
{
	GMimeContentFileData * fd = (GMimeContentFileData*) content->data;
	g_return_if_fail (fd!=NULL);

	g_mime_content_file_init_filename (fd, "w+");
}

static void
g_mime_content_file_append (GMimeContent* content, const guchar * bytes, guint byte_qty)
{
	GMimeContentFileData * fd = (GMimeContentFileData*) content->data;
	g_return_if_fail (fd!=NULL);
	g_return_if_fail (fd->fp!=NULL);

	fwrite (bytes, 1, byte_qty, fd->fp);
	content->len += byte_qty;
}

static void
g_mime_content_file_append_pre_encoded (GMimeContent* content,
                                        const guchar * encoded,
                                        guint encoded_len,
                                        GMimePartEncodingType encoding)
{
	guint dec_len = 0;
	guchar * dec = NULL;
	GMimeContentFileData * data = NULL;

	g_return_if_fail (content!=NULL);
	g_return_if_fail (content->data!=NULL);
	g_return_if_fail (encoded!=NULL);
	g_return_if_fail (encoded_len>0);

	data = (GMimeContentFileData*) content->data;
	g_return_if_fail (data->fp!=NULL);

	g_mime_content_append_encoded_impl (content, encoded, encoded_len, encoding,
	                                    &dec, &dec_len);
	fwrite (dec, 1, dec_len, data->fp);
	content->len += dec_len;
	g_free (dec);
}

static void
g_mime_content_file_set_end (GMimeContent* content)
{
	GMimeContentFileData * data;

	g_return_if_fail (content!=NULL);
	g_return_if_fail (content->data!=NULL);

	data = (GMimeContentFileData*) content->data;
	g_return_if_fail (data->fp!=NULL);
	fclose (data->fp);
	data->fp = NULL;
}


static gint
g_mime_content_file_iterator_step (GMimeContentIterator* it, guchar * buf, guint buflen)
{
	g_return_val_if_fail (it!=NULL, 0);
	g_return_val_if_fail (it->data!=NULL, 0);
	g_return_val_if_fail (buf!=NULL, 0);
	g_return_val_if_fail (buflen>0, 0);

	return fread (buf, 1, buflen, (FILE*)it->data);
}

static gboolean
g_mime_content_file_iterator_readln (GMimeContentIterator  * it, GByteArray * setme)
{
	gchar buf[2048] = { '\0' };
	
	g_return_val_if_fail (it!=NULL, FALSE);
	g_return_val_if_fail (it->data!=NULL, FALSE);
	g_return_val_if_fail (setme!=NULL, FALSE);

	g_byte_array_set_size (setme, 0);
	if (!fgets (buf, sizeof(buf), (FILE*)it->data))
		return FALSE;
	g_byte_array_append (setme, buf, strlen(buf)+1);
	return TRUE;
}


static void
g_mime_content_file_iterator_destroy (GMimeContentIterator* it)
{
	FILE * fp;

	g_return_if_fail (it!=NULL);

	if ((fp = (FILE*) it->data))
		fclose (fp);

	g_free (it);
}

static GMimeContentIterator*
g_mime_content_file_iterator_new (const GMimeContent * content)
{
	GMimeContentIterator * it;
	GMimeContentFileData * data;

	/* sanity checks */
	data = (GMimeContentFileData*) content->data;
	g_return_val_if_fail (data!=NULL, NULL);
	g_return_val_if_fail (data->filename!=NULL, NULL);

	it = g_new (GMimeContentIterator, 1);
	it->step_func = g_mime_content_file_iterator_step;
	it->readln_func = g_mime_content_file_iterator_readln;
	it->destroy_func = g_mime_content_file_iterator_destroy;
	it->data = fopen (data->filename, "r");
	return it;
}

GMimeContent*
g_mime_content_file_new (void)
{
	GMimeContent * content = g_new0 (GMimeContent, 1);

	content->content_set_begin_func = g_mime_content_file_set_begin;
	content->content_append_func = g_mime_content_file_append;
	content->content_append_pre_encoded_func = g_mime_content_file_append_pre_encoded;
	content->content_set_end_func =g_mime_content_file_set_end;

	content->content_get_iterator_func = g_mime_content_file_iterator_new;

	content->data = g_new0 (GMimeContentFileData, 1);
	content->data_destroy = g_mime_content_file_data_destroy;

	return content;
}

/**
***  Memory
**/

static void
g_mime_content_array_data_destroy (gpointer data)
{
	g_return_if_fail (data!=NULL);

	g_byte_array_free ((GByteArray*)data, TRUE);
}

static void
g_mime_content_array_set_begin (GMimeContent * content)
{
	/* Nothing to do */
}

static void
g_mime_content_array_append (GMimeContent * content,
                             const guchar * data,
                             guint data_len)
{
	GByteArray * bytes = (GByteArray*) content->data;

	g_byte_array_append (bytes, data, data_len);
	content->len += data_len;
}

static void
g_mime_content_array_append_pre_encoded (GMimeContent * content,
                                         const guchar * data,
                                         guint data_len,
                                         GMimePartEncodingType encoding)
{
	gint dec_len = 0;
	guchar * dec = NULL;
	GByteArray * bytes = (GByteArray*) content->data;

	g_mime_content_append_encoded_impl (content, data, data_len, encoding, &dec, &dec_len);
	g_byte_array_append (bytes, dec, dec_len);
	content->len += dec_len;
	g_free (dec);
}

static void
g_mime_content_array_set_end (GMimeContent * content)
{
	/* Nothing to do */
}

typedef struct
{
	GByteArray * bytes;
	guint read_pos;
}
ArrayIteratorData;

static gint
g_mime_content_array_iterator_step (GMimeContentIterator * it, guchar * buf, guint buflen)
{
	gint len_read;
	gint remaining;
	ArrayIteratorData * aid;

	/* sanity checks */
	g_return_val_if_fail (it!=NULL, 0);
	g_return_val_if_fail (it->data!=NULL, 0);
	g_return_val_if_fail (buf!=NULL, 0);
	g_return_val_if_fail (buflen>0, 0);
	aid = (ArrayIteratorData*) it->data;

	if (aid->read_pos >= aid->bytes->len)
		return 0;

	remaining = aid->bytes->len - aid->read_pos;
	len_read = MIN (remaining, buflen);
	memcpy (buf, aid->bytes->data+aid->read_pos, len_read);
	aid->read_pos += len_read;

	return len_read;
}

static gboolean
g_mime_content_array_iterator_readln (GMimeContentIterator * it, GByteArray * setme)
{
	const guchar * base;
	const guchar * pch;
	ArrayIteratorData * aid;

	/* sanity checks */
	g_return_val_if_fail (it!=NULL, 0);
	g_return_val_if_fail (setme!=NULL, 0);
	g_return_val_if_fail (setme!=NULL, 0);
	aid = (ArrayIteratorData*) it->data;

	if (aid->read_pos >= aid->bytes->len)
		return FALSE;

	g_byte_array_set_size (setme, 0);
	base = aid->bytes->data + aid->read_pos;
	pch = strchr (base, '\n');
	if (pch != NULL) {
		gint len = pch+1-base; /* incldue '\n' */
		g_byte_array_append (setme, base, len);
		aid->read_pos += len;
	} else {
		gint len = aid->bytes->len - aid->read_pos;
		g_byte_array_append (setme, base, len);
		aid->read_pos += len;
	}
	g_byte_array_append (setme, "", 1); /* '\0' */
	return setme->len!=1;
}

static void
g_mime_content_array_iterator_destroy (GMimeContentIterator * it)
{
	g_free (it->data);
	g_free (it);
}

/**
***  Public interface -- Life Cycle
**/

static GMimeContentIterator*
g_mime_content_array_iterator_new (const GMimeContent * content)
{
	ArrayIteratorData * aid;
	GMimeContentIterator * it;

	aid = g_new (ArrayIteratorData, 1);
	aid->bytes = (GByteArray*) content->data;
	aid->read_pos = 0;

	it = g_new (GMimeContentIterator, 1);
	it->step_func = g_mime_content_array_iterator_step;
	it->readln_func = g_mime_content_array_iterator_readln;
	it->destroy_func = g_mime_content_array_iterator_destroy;
	it->data = aid;
	return it;
}

GMimeContent*
g_mime_content_array_new (void)
{
	GMimeContent * content = g_new0 (GMimeContent, 1);

	content->content_set_begin_func = g_mime_content_array_set_begin;
	content->content_append_func = g_mime_content_array_append;
	content->content_append_pre_encoded_func = g_mime_content_array_append_pre_encoded;
	content->content_set_end_func =g_mime_content_array_set_end;
	content->content_get_iterator_func = g_mime_content_array_iterator_new;
	content->data = g_byte_array_new ();
	content->data_destroy = g_mime_content_array_data_destroy;

	return content;
}

void
g_mime_content_destroy (GMimeContent * c)
{
	g_return_if_fail (c!=NULL);

	(*c->data_destroy)(c->data);
}

/**
***  Public interface -- Set Contents
**/

void
g_mime_content_set_begin (GMimeContent * c)
{
	g_return_if_fail (c!=NULL);

	(c->content_set_begin_func)(c);
}

void
g_mime_content_append (GMimeContent * c, const guchar * data, guint data_len)
{
	g_return_if_fail (c!=NULL);
	g_return_if_fail (data!=NULL);
	g_return_if_fail (data_len>0);

	(*c->content_append_func)(c, data, data_len);
}

void
g_mime_content_append_pre_encoded (GMimeContent * c,
                                       const guchar * encoded,
                                       guint encoded_len,
                                       GMimePartEncodingType encoding)
{
	g_return_if_fail (c!=NULL);
	g_return_if_fail (encoded!=NULL);
	g_return_if_fail (encoded_len>0);

	(*c->content_append_pre_encoded_func)(c, encoded, encoded_len, encoding);
}

void
g_mime_content_set_end (GMimeContent * c)
{
	g_return_if_fail (c!=NULL);

	(*c->content_set_end_func)(c);
}

void
g_mime_content_set_all (GMimeContent* c, const guchar* buf, guint buflen)
{
	g_return_if_fail (c!=NULL);
	g_return_if_fail (buf!=NULL);
	g_return_if_fail (buflen>0);

	g_mime_content_set_begin (c);
	g_mime_content_append (c, buf, buflen);
	g_mime_content_set_end (c);
}

void
g_mime_content_set_all_pre_encoded (GMimeContent* c,
                                    const guchar* buf,
                                    guint buflen,
                                    GMimePartEncodingType type)
{
	g_return_if_fail (c!=NULL);
	g_return_if_fail (buf!=NULL);
	g_return_if_fail (buflen>0);

	g_mime_content_set_begin (c);
	g_mime_content_append_pre_encoded (c, buf, buflen, type);
	g_mime_content_set_end (c);
}

/**
***  Public interface -- Get Contents
**/
 
void
g_mime_content_get_all (const GMimeContent * c, GByteArray * setme)
{
	GMimeContentIterator * it;

	g_return_if_fail (c!=NULL);
	g_return_if_fail (setme!=NULL);

	g_byte_array_set_size (setme, c->len);
	it = g_mime_content_get_iterator (c);
	g_mime_content_iterator_step (it, setme->data, setme->len);
	g_mime_content_iterator_destroy (it);
	g_byte_array_append (setme, "", 1); /* for trailing '\0' */
}

/**
 * g_mime_content_get_iterator: Get an Iterator for reading GMimeContent.
 * @content: the content whose data we want to read through
 *
 * This function returnsa a read Iterator for walking through
 * the specified content object's content information.
 */
GMimeContentIterator*
g_mime_content_get_iterator (const GMimeContent * content)
{
	g_return_val_if_fail (content!=NULL, NULL);

	return (*content->content_get_iterator_func)(content);
}

/**
 * g_mime_content_iterator_step: Read a block of content data from a GMimeContentIterator.
 * @iterator: The GMimeContentIterator to read from.
 * @buf: the buf to store the read data in.
 * @maxlen: the maximum number of bytes to write to the buf array.
 *
 * This function is like fread() for a GMimeContentIterator and returns
 * the number of bytes read into buf.
 */
gint
g_mime_content_iterator_step (GMimeContentIterator  * it, guchar * buf, guint maxlen)
{
	g_return_val_if_fail (it!=NULL, 0);
	g_return_val_if_fail (buf!=NULL, 0);
	g_return_val_if_fail (maxlen>0, 0);

	return (*it->step_func)(it, buf, maxlen);
}

/**
 * g_mime_content_iterator_readln: Read a line from GMimeContentIterator.
 * @iterator: The iterator to read from
 * @line: a GByteArray to be filled with a zero-terminated line.
 *
 * Reads a single line from a GMimeContentIterator and places the line,
 * with a zero terminator, in the user-specified line buffer.  This function
 * returns TRUE if a line was successfully read, FALSE otherwise.
 */
gboolean
g_mime_content_iterator_readln (GMimeContentIterator * it, GByteArray * line)
{
	g_return_val_if_fail (it!=NULL, FALSE);
	g_return_val_if_fail (line!=NULL, FALSE);

	return (*it->readln_func)(it, line);
}

/**
 * g_mime_content_iterator_destroy: Destroy a GMimeContentIterator object
 * @iterator: The iterator to destroy
 *
 * Destroys an iterator after you're done reading the contents of a
 * GMimeContent object with it.
 */
void
g_mime_content_iterator_destroy (GMimeContentIterator * it)
{
	g_return_if_fail (it!=NULL);

	(*it->destroy_func)(it);
}

/**
***  Public interface -- Encoding
**/

/**
 * g_mime_content_set_encoding: Set the content encoding
 * @mime_content: Mime content
 * @encoding: Mime encoding
 *
 * Set the content encoding for the specified content. Available
 * values for the encoding are: GMIME_PART_ENCODING_DEFAULT,
 * GMIME_PART_ENCODING_7BIT, GMIME_PART_ENCODING_8BIT,
 * GMIME_PART_ENCODING_BASE64 and GMIME_PART_ENCODING_QUOTEDPRINTABLE.
 **/
void
g_mime_content_set_encoding (GMimeContent *content, GMimePartEncodingType encoding)
{
	g_return_if_fail (content != NULL);

	content->encoding = encoding;
}


/**
 * g_mime_content_get_encoding: Get the content encoding
 * @mime_content: Mime content
 *
 * Returns the content encoding for the specified content. The
 * return value will be one of the following:
 * GMIME_PART_ENCODING_DEFAULT, GMIME_PART_ENCODING_7BIT,
 * GMIME_PART_ENCODING_8BIT, GMIME_PART_ENCODING_BASE64 or
 * GMIME_PART_ENCODING_QUOTEDPRINTABLE.
 **/
GMimePartEncodingType
g_mime_content_get_encoding (GMimeContent *content)
{
	g_return_val_if_fail (content != NULL, GMIME_PART_ENCODING_DEFAULT);
	
	return content->encoding;
}
